from abc import ABCMeta
from abc import abstractmethod

from .oids import Oids
from .digest_algorithm import DigestAlgorithm


class SignatureAlgorithms(object):
    MD5_WITH_RSA = 'MD5WithRSA'
    SHA1_WITH_RSA = 'SHA1WithRSA'
    SHA256_WITH_RSA = 'SHA256WithRSA'
    SHA384_WITH_RSA = 'SHA384WithRSA'
    SHA512_WITH_RSA = 'SHA512WithRSA'


class SignatureAlgorithm(object):
    MD5_WITH_RSA = None
    SHA1_WITH_RSA = None
    SHA256_WITH_RSA = None
    SHA384_WITH_RSA = None
    SHA512_WITH_RSA = None

    def __init__(self, name, oid, xml_uri, digest_algorithm, pk_algorithm):
        self.__name = name
        self.__oid = oid
        self.__xml_uri = xml_uri
        self.__digest_algorithm = digest_algorithm
        self.__pk_algorithm = pk_algorithm

    def __eq__(self, instance):
        if instance is None:
            return False

        return self.__oid == instance.oid

    @staticmethod
    def _algorithms():
        return [
            SignatureAlgorithm.MD5_WITH_RSA,
            SignatureAlgorithm.SHA1_WITH_RSA,
            SignatureAlgorithm.SHA256_WITH_RSA,
            SignatureAlgorithm.SHA384_WITH_RSA,
            SignatureAlgorithm.SHA512_WITH_RSA
        ]

    @staticmethod
    def _safe_algorithms():
        return [
            SignatureAlgorithm.SHA1_WITH_RSA,
            SignatureAlgorithm.SHA256_WITH_RSA,
            SignatureAlgorithm.SHA384_WITH_RSA,
            SignatureAlgorithm.SHA512_WITH_RSA
        ]

    @staticmethod
    def get_instance_by_name(name):
        filtered_list = list(filter(lambda s: s.name == name,
                                    SignatureAlgorithm._algorithms()))
        alg = filtered_list[0] if len(filtered_list) > 0 else None

        if alg is None:
            raise Exception('Unrecognized signature algorithm name: %s' % name)

        return alg

    @staticmethod
    def get_instance_by_oid(oid):
        filtered_list = list(filter(lambda s: s.oid == oid,
                                    SignatureAlgorithm._algorithms()))
        alg = filtered_list[0] if len(filtered_list) > 0 else None

        if alg is None:
            raise Exception('Unrecognized signature algorithm oid: %s' % oid)

        return alg

    @staticmethod
    def get_instance_by_xml_uri(xml_uri):
        filtered_list = list(filter(lambda s: s.xml_uri == xml_uri,
                                    SignatureAlgorithm._algorithms()))
        alg = filtered_list[0] if len(filtered_list) > 0 else None

        if alg is None:
            raise Exception('Unrecognized signature algorithm XML URI: %s',
                            xml_uri)

        return alg

    @staticmethod
    def get_instance_by_api_model(algorithm):
        if algorithm == SignatureAlgorithms.MD5_WITH_RSA:
            return SignatureAlgorithm.MD5_WITH_RSA
        elif algorithm == SignatureAlgorithms.SHA1_WITH_RSA:
            return SignatureAlgorithm.SHA1_WITH_RSA
        elif algorithm == SignatureAlgorithms.SHA256_WITH_RSA:
            return SignatureAlgorithm.SHA256_WITH_RSA
        elif algorithm == SignatureAlgorithms.SHA384_WITH_RSA:
            return SignatureAlgorithm.SHA384_WITH_RSA
        elif algorithm == SignatureAlgorithms.SHA512_WITH_RSA:
            return SignatureAlgorithm.SHA512_WITH_RSA
        else:
            raise Exception('Unsupported signature algorithm: %s' % algorithm)

    @property
    def name(self):
        return self.__name

    @name.setter
    def name(self, value):
        self.__name = value

    @property
    def oid(self):
        return self.__oid

    @oid.setter
    def oid(self, value):
        self.__oid = value

    @property
    def xml_uri(self):
        return self.__xml_uri

    @xml_uri.setter
    def xml_uri(self, value):
        self.__xml_uri = value

    @property
    def digest_algorithm(self):
        return self.__digest_algorithm

    @digest_algorithm.setter
    def digest_algorithm(self, value):
        self.__digest_algorithm = value

    @property
    def pk_algorithm(self):
        return self.__pk_algorithm

    @pk_algorithm.setter
    def pk_algorithm(self, value):
        self.__pk_algorithm = value


class PKAlgorithms(object):
    RSA = 'RSA'


class PKAlgorithm(object):
    __metaclass__ = ABCMeta

    RSA = None

    def __init__(self, name, oid):
        self.__name = name
        self.__oid = oid

    def __eq__(self, instance):
        if instance is None:
            return False

        return self.__oid == instance.oid

    @staticmethod
    def _algorithms():
        return [PKAlgorithm.RSA]

    @staticmethod
    def get_instance_by_name(name):
        filtered_list = list(filter(lambda p: p.name == name,
                                    PKAlgorithm._algorithms()))
        alg = filtered_list[0] if len(filtered_list) > 0 else None

        if alg is None:
            raise Exception('Unrecognized private key algorithm name: %s' %
                            name)

        return alg

    @staticmethod
    def get_instance_by_oid(oid):
        filtered_list = list(filter(lambda p: p.oid == oid,
                                    PKAlgorithm._algorithms()))
        alg = filtered_list[0] if len(filtered_list) > 0 else None

        if alg is None:
            raise Exception('Unrecognized private key algorithm oid: %s' % oid)

        return alg

    @staticmethod
    def get_instance_api_model(algorithm):
        if algorithm is PKAlgorithms.RSA:
            return PKAlgorithm.RSA
        else:
            raise Exception('Unsupported private key algorithm: %s' % algorithm)

    @property
    def name(self):
        return self.__name

    @name.setter
    def name(self, value):
        self.__name = value

    @property
    def oid(self):
        return self.__oid

    @oid.setter
    def oid(self, value):
        self.__oid = value

    @abstractmethod
    def get_signature_algorithm(self, digest_algorithm):
        pass


class RSASignatureAlgorithm(SignatureAlgorithm):

    def __init__(self, digest_algorithm):
        name = '%s with RSA' % digest_algorithm
        pk_algorithm = PKAlgorithm.RSA

        if digest_algorithm == DigestAlgorithm.MD5:
            xml_uri = 'http://www.w3.org/2001/04/xmldsig-more#rsa-md5'
            oid = Oids.MD5_WITH_RSA
        elif digest_algorithm == DigestAlgorithm.SHA1:
            xml_uri = 'http://www.w3.org/2000/09/xmldsig#rsa-sha1'
            oid = Oids.SHA1_WITH_RSA
        elif digest_algorithm == DigestAlgorithm.SHA256:
            xml_uri = 'http://www.w3.org/2001/04/xmldsig-more#rsa-sha256'
            oid = Oids.SHA256_WITH_RSA
        elif digest_algorithm == DigestAlgorithm.SHA384:
            xml_uri = 'http://www.w3.org/2001/04/xmldsig-more#rsa-sha384'
            oid = Oids.SHA384_WITH_RSA
        elif digest_algorithm == DigestAlgorithm.SHA512:
            xml_uri = 'http://www.w3.org/2001/04/xmldsig-more#rsa-sha512'
            oid = Oids.SHA512_WITH_RSA
        else:
            raise Exception('Unsupported digest algorithms: %s' %
                            digest_algorithm)

        super(RSASignatureAlgorithm, self).__init__(name, oid, xml_uri,
                                                    digest_algorithm,
                                                    pk_algorithm)


SignatureAlgorithm.MD5_WITH_RSA = RSASignatureAlgorithm(DigestAlgorithm.MD5)
SignatureAlgorithm.SHA1_WITH_RSA = RSASignatureAlgorithm(DigestAlgorithm.SHA1)
SignatureAlgorithm.SHA256_WITH_RSA = \
    RSASignatureAlgorithm(DigestAlgorithm.SHA256)
SignatureAlgorithm.SHA384_WITH_RSA = \
    RSASignatureAlgorithm(DigestAlgorithm.SHA384)
SignatureAlgorithm.SHA512_WITH_RSA = \
    RSASignatureAlgorithm(DigestAlgorithm.SHA512)


class RSAPKAlgorithm(PKAlgorithm):

    def __init__(self):
        name = 'RSA'
        oid = Oids.RSA
        super(RSAPKAlgorithm, self).__init__(name, oid)

    def get_signature_algorithm(self, digest_algorithm):
        return RSASignatureAlgorithm(digest_algorithm)


PKAlgorithm.RSA = RSAPKAlgorithm()

__all__ = [
    'SignatureAlgorithms',
    'SignatureAlgorithm',
    'PKAlgorithms',
    'PKAlgorithm',
    'RSASignatureAlgorithm',
    'RSAPKAlgorithm'
]



from .pdf_mark_element_type import PdfMarkElementType
from .pdf_mark_element import PdfMarkElement


class PdfMarkQRCodeElement(PdfMarkElement):

    def __init__(self, relative_container=None, qr_code_data=None):
        super(PdfMarkQRCodeElement, self).__init__(PdfMarkElementType.QR_CODE,
                                                   relative_container)

        self.__qr_code_data = qr_code_data
        self.__draw_quiet_zones = False

    @property
    def qr_code_data(self):
        return self.__qr_code_data

    @qr_code_data.setter
    def qr_code_data(self, value):
        self.__qr_code_data = value

    # region FluentAPI

    def with_qr_code_data(self, qr_code_data):
        self.__qr_code_data = qr_code_data
        return self

    def draw_quiet_zones(self):
        self.__draw_quiet_zones = True
        return self

    # endregion

    def to_model(self):
        model = super(PdfMarkQRCodeElement, self).to_model()
        model['qrCodeData'] = self.__qr_code_data
        model['qrCddeDrawQuietZones'] = self.__draw_quiet_zones
        return model


__all__ = ['PdfMarkQRCodeElement']



from .apis import Apis
from .rest_pki_client import _get_api_version
from .signature_finisher import SignatureFinisher


class PadesSignatureFinisher(SignatureFinisher):

    def _check_compatibility(self):
        api_version = _get_api_version(self._client, Apis.COMPLETE_PADES)
        if api_version < 2:
            raise Exception('The PadesSignatureFinisher class can only be used'
                            ' with REST PKI 1.11 or later.')

    def _get_api_route(self):
        return 'Api/v2/PadesSignatures/%s/SignedBytes' % self._token


__all__ = ['PadesSignatureFinisher']



from .digest_algorithm_and_value import DigestAlgorithmAndValue


class SignaturePolicyIdentifier(object):

    def __init__(self, model):
        self.__digest = None
        digest = model.get('digest', None)
        if digest is not None:
            self.__digest = DigestAlgorithmAndValue(digest)

        self.__oid = model.get('oid', None)
        self.__uri = model.get('uri', None)

    @property
    def digest(self):
        return self.__digest

    @digest.setter
    def digest(self, value):
        self.__digest = value

    @property
    def oid(self):
        return self.__oid

    @oid.setter
    def oid(self, value):
        self.__oid = value

    @property
    def uri(self):
        return self.__uri

    @uri.setter
    def uri(self, value):
        self.__uri = value


__all__ = ['SignaturePolicyIdentifier']



from .resource_content_or_reference import ResourceContentOrReference


class PdfMarkImage(object):

    def __init__(self, image_content=None, mime_type=None):
        self.__resource = ResourceContentOrReference()
        self.__resource.mime_type = mime_type
        self.__resource.content = image_content

    @property
    def resource(self):
        return self.__resource

    @resource.setter
    def resource(self, value):
        self.__resource = value

    def to_model(self):
        return {
            'resource': self.__resource.to_model()
            if self.__resource is not None else None
        }


__all__ = ['PdfMarkImage']



from .pades_size import PadesSize
from .page_orientations import PageOrientations
from .paper_sizes import PaperSizes


class PadesPageOptimization(object):

    def __init__(self, paper_size=None, custom_paper_size=None):
        if custom_paper_size is not None:
            self.__paper_size = paper_size
        else:
            self.__paper_size = PaperSizes.CUSTOM
            self.__custom_paper_size = custom_paper_size
        self.__page_orientation = PageOrientations.AUTO

    # region "paper_size" accessors

    @property
    def paper_size(self):
        return self.__get_paper_size()

    def __get_paper_size(self):
        return self.__paper_size

    @paper_size.setter
    def paper_size(self, value):
        self.__set_paper_size(value)

    def __set_paper_size(self, value):
        if value is None:
            raise Exception('The provided "paper_size" is not valid')
        self.__paper_size = value

    # endregion

    # region "custom_paper_size" accessors

    @property
    def custom_paper_size(self):
        return self.__get_custom_paper_size()

    def __get_custom_paper_size(self):
        return self.__custom_paper_size

    @custom_paper_size.setter
    def custom_paper_size(self, value):
        self.__set_custom_paper_size(value)

    def __set_custom_paper_size(self, value):
        if value is None:
            raise Exception('The provided "custom_paper_size" is not valid')
        self.__custom_paper_size = value

    # endregion

    # region "page_orientation" accessors

    @property
    def page_orientation(self):
        return self.__get_page_orientation()

    def __get_page_orientation(self):
        return self.__page_orientation

    @page_orientation.setter
    def page_orientation(self, value):
        self.__set_page_orientation(value)

    def __set_page_orientation(self, value):
        if value is None:
            raise Exception('The provided "page_orientation" is not valid')
        self.__page_orientation = value

    # endregion

    def to_model(self):
        return {
            'paperSize': self.__paper_size,
            'customPaperSize': self.__custom_paper_size.to_model()
                               if self.__custom_paper_size is not None
                               else None,
            'pageOrientation': self.__page_orientation
        }

    @staticmethod
    def create_from_model(model):
        entity = PadesPageOptimization()
        entity.paper_size = model.get('paperSize', None)
        custom_paper_size = model.get('customPaperSize', None)
        if custom_paper_size is not None:
            entity.custom_paper_size = \
                PadesSize.create_from_model(custom_paper_size)
        page_orientation = model.get('pageOrientation', None)
        entity.page_orientation = page_orientation or PageOrientations.AUTO
        return entity


__all__ = ['PadesPageOptimization']



from .pk_certificate import PKCertificate
from .validation import ValidationResults


class AuthenticationResult(object):

    def __init__(self, model):
        self.__certificate = None
        certificate = model.get('certificate', None)
        if certificate is not None:
            self.__certificate = PKCertificate(certificate)

        self.__validation_results = None
        validation_results = model.get('validationResults', None)
        if validation_results is not None:
            self.__validation_results = ValidationResults(validation_results)

    @property
    def certificate(self):
        return self.__certificate

    @certificate.setter
    def certificate(self, value):
        self.__certificate = value

    @property
    def validation_results(self):
        return self.__validation_results

    @validation_results.setter
    def validation_results(self, value):
        self.__validation_results = value


__all__ = ['AuthenticationResult']



class NamespaceManager(object):

    def __init__(self):
        self.__namespaces = []

    @property
    def namespaces(self):
        return self.__namespaces

    @namespaces.setter
    def namespaces(self, value):
        self.__namespaces = value

    def add_namespace(self, prefix, uri):
        if prefix is None or uri is None:
            raise Exception('Prefix and uri parameters cannot be None')
        self.__namespaces.append({'prefix': prefix, 'uri': uri})


__all__ = ['NamespaceManager']
